from pathlib import Path
from typing import Iterator

import polars as pl

from .exceptions import ParsingError


def roots(orca_output: Path) -> pl.DataFrame:
    """Parse roots from ORCA output into a Polars DataFrame."""
    # Read file content
    content = Path(orca_output).read_text().splitlines()

    # Find the start of root section(s)
    start_indicies = []
    for line in content:
        if "TD-DFT/TDA EXCITED STATES" in line or "TD-DFT EXCITED STATES" in line:
            start_indicies.append(content.index(line))
            # Find the custom offset (different for different tables..)
            while not content[start_indicies[-1]].startswith("STATE"):
                start_indicies[-1] += 1

    # Parse roots
    roots = []
    for start_idx in start_indicies:
        for root_lines in _iter_roots(content[start_idx:]):
            root_data = _parse_single_root(root_lines)
            roots.append(root_data)

    # Flatten the data structure
    flattened_roots = []
    for root in roots:
        for orb_contrib in root["orbital_transitions"]:
            flattened_roots.append(
                {
                    "root": root["root"],
                    "mult": root["mult"],
                    "donor": orb_contrib["donor"],
                    "acceptor": orb_contrib["acceptor"],
                    "weight": orb_contrib["weight"],
                    "energy_cm": root["energy_cm"],
                }
            )

    # Create and transform DataFrame
    df = pl.DataFrame(flattened_roots)

    if len(df) == 0:
        raise ParsingError("No roots found in input")

    return df.sort("root", "weight", descending=[False, True])


def _parse_single_root(root_lines: list[str]) -> dict:
    """Parse a single root block into a dictionary."""
    # Parse header line
    header = root_lines[0].strip()
    state_num = int(header[5 : header.index(":")])
    parts = header.strip().split()
    energy_ev = float(parts[5])
    energy_cm = float(parts[7])
    spin_projection = float(parts[11])
    mult = int(parts[13])

    # Parse root contributions
    orbital_transitions = []
    for line in root_lines[1:]:
        parts = line.replace("->", "").replace(":", "").replace(")", "").strip().split()
        orbital_transitions.append(
            {
                "donor": str(parts[0]),
                "acceptor": str(parts[1]),
                "weight": float(parts[2]),
            }
        )

    return {
        "root": state_num,
        "energy_ev": energy_ev,
        "energy_cm": energy_cm,
        "spin_projection": spin_projection,
        "mult": mult,
        "orbital_transitions": orbital_transitions,
    }


def _iter_roots(lines: list[str]) -> Iterator[list[str]]:
    """Iterate over root blocks in the input text."""
    current_state = []

    for i, line in enumerate(lines):
        line = line.strip()
        if not line:
            if lines[i + 1].strip().startswith("STATE"):
                continue
            else:
                break

        if line.startswith("STATE"):
            if current_state:
                yield current_state
            current_state = [line]
        elif current_state:
            current_state.append(line)

    if current_state:  # Don't forget the last state
        yield current_state



from pathlib import Path

import polars as pl

from .utils import find_table_start


def ir_spectrum(orca_output: Path) -> pl.DataFrame:
    lines = Path(orca_output).read_text().splitlines()

    TABLE_HEADER = "IR SPECTRUM"
    TABLE_HEADER_OFFSET = 6

    table_start_idx = find_table_start(lines, TABLE_HEADER, TABLE_HEADER_OFFSET)

    # Collect table
    rows = []
    for row in lines[table_start_idx:]:
        # Stop on empty line
        if not row.strip():
            break

        rows.append(row)

    processed_rows = [
        row.replace(":", "").replace("(", "").replace(")", "").strip().split()
        for row in rows
    ]

    df = pl.DataFrame(
        processed_rows,
        orient="row",
        schema={
            "mode": pl.Int64,
            "frequency_cm": pl.Float64,
            "epsilon": pl.Float64,
            "intensity": pl.Float64,
            "t2": pl.Float64,
            "tx": pl.Float64,
            "ty": pl.Float64,
            "tz": pl.Float64,
        },
    )

    df = df.with_columns(
        (pl.col("intensity") / pl.col("intensity").max()).alias("rel_intensity")
    )

    return df



import altair as alt
import polars as pl

from blackfish.charts.gaussian_broadening import apply_gaussian_filter


def soc_absorption_spectrum_chart(df: pl.DataFrame, fwhm: int = 2000) -> alt.Chart:
    bar = (
        alt.Chart(df)
        .mark_bar(opacity=0.5)
        .encode(
            x=alt.X("energy_cm:Q", title="Energy [1/cm]"),
            y=alt.Y("rel_intensity:Q", title="Rel. Intensity"),
            color=alt.value("red"),
            tooltip=["energy_cm", "rel_intensity"],
        )
    )

    simulated_df = apply_gaussian_filter(df, "energy_cm", "rel_intensity", fwhm, 10)

    line = (
        alt.Chart(simulated_df)
        .mark_line()
        .encode(x="energy_cm:Q", y="rel_intensity:Q", color=alt.value("black"))
    )
    chart = bar + line

    return chart



from typing import List

from blackfish.parsing.exceptions import ParsingError


def find_table_start(lines: List[str], header: str, offset: int) -> int:
    """
    Find the starting index of a table in ORCA output files.

    Args:
        lines: List of strings containing the file content
        header: The header text to search for
        offset: Number of lines to skip after the header (default=5)

    Returns:
        int: The index where the table data starts

    Raises:
        ParsingError: If the header is not found in the file
    """
    try:
        table_start_idx = next(
            i for i, line in enumerate(lines) if header in line.strip()
        )
    except StopIteration:
        raise ParsingError(f"Could not find '{header}'")

    return table_start_idx + offset



from pathlib import Path

import polars as pl

from .utils import find_table_start


def nacme(orca_output: Path) -> pl.DataFrame:
    lines = Path(orca_output).read_text().splitlines()

    TABLE_HEADER = "CARTESIAN NON-ADIABATIC COUPLINGS"
    TABLE_HEADER_OFFSET = 5

    table_start_idx = find_table_start(lines, TABLE_HEADER, TABLE_HEADER_OFFSET)

    # Collect table
    rows = []
    for row in lines[table_start_idx:]:
        # Stop on empty line
        if not row.strip():
            break

        rows.append(row.split())

    df = pl.DataFrame(
        rows,
        schema={"id": int, "symbol": str, "_": str, "x": float, "y": float, "z": float},
        orient="row",
    ).drop("_")

    # Compute magnitude
    df = df.with_columns(
        (pl.col("x").abs() + pl.col("y").abs() + pl.col("z").abs()).alias("magnitude")
    )

    return df



from pathlib import Path

import polars as pl

from .exceptions import ParsingError
from .utils import find_table_start


def soc_absorption_spectrum(orca_output: Path) -> pl.DataFrame:
    lines = Path(orca_output).read_text().splitlines()

    TABLE_HEADER = (
        "SOC CORRECTED ABSORPTION SPECTRUM VIA TRANSITION ELECTRIC DIPOLE MOMENTS"
    )
    TABLE_HEADER_OFFSET = 5

    table_start_idx = find_table_start(lines, TABLE_HEADER, TABLE_HEADER_OFFSET)

    # Collect table
    rows = []
    for row in lines[table_start_idx:]:
        # Stop on empty line
        if not row.strip():
            break

        rows.append(row)

    if not rows:
        raise ParsingError("No data found in SOC absorption spectrum table")

    # Process table
    processed_rows = []
    for row in rows:
        row = row.replace("A", "").replace("B", "").replace("->", "")
        to_state, to_spin = row.split()[1].split("-")
        processed_row = [to_state, to_spin] + row.split()[2:]
        processed_rows.append(processed_row)

    df = pl.DataFrame(
        processed_rows,
        orient="row",
        schema={
            "state": pl.Int64,
            "mult": pl.Float64,
            "energy_ev": pl.Float64,
            "energy_cm": pl.Float64,
            "wavelength_nm": pl.Float64,
            "osc_strength": pl.Float64,
            "d2": pl.Float64,
            "dx": pl.Float64,
            "dy": pl.Float64,
            "dz": pl.Float64,
        },
    )

    df = df.with_columns(
        (pl.col("osc_strength") / pl.col("osc_strength").max()).alias("rel_intensity")
    )

    return df



from pathlib import Path

import polars as pl

import blackfish as bf


class ORCA:
    def __init__(self, output_file: Path):
        self.output_file = output_file

    @property
    def ir_spectrum(self) -> pl.DataFrame:
        return bf.ir_spectrum(self.output_file)

    @property
    def nacme(self) -> pl.DataFrame:
        return bf.nacme(self.output_file)

    @property
    def roots(self) -> pl.DataFrame:
        return bf.roots(self.output_file)

    @property
    def soc_absorption_spectrum(self) -> pl.DataFrame:
        return bf.soc_absorption_spectrum(self.output_file)

    @property
    def soc_states(self) -> pl.DataFrame:
        return bf.soc_states(self.output_file)

    @property
    def soc_absorption_spectrum_chart(self) -> None:
        return bf.soc_absorption_spectrum_chart(self.soc_absorption_spectrum)
